import os
import json


class JsonStatSchema:
    def __init__(self):
        filename = os.path.join(os.path.dirname(
            __file__), "schemas", "jsonstat.json")
        with open(filename) as f:
            self.__all = json.loads(f.read())

        filename = os.path.join(os.path.dirname(
            __file__), "schemas", "collection.json")
        with open(filename) as f:
            self.__collection = json.loads(f.read())

        filename = os.path.join(os.path.dirname(
            __file__), "schemas", "dataset.json")
        with open(filename) as f:
            self.__dataset = json.loads(f.read())

        filename = os.path.join(os.path.dirname(
            __file__), "schemas", "dimension.json")
        with open(filename) as f:
            self.__dimension = json.loads(f.read())

    @property
    def dimension(self):
        return self.__dimension

    @property
    def dataset(self):
        return self.__dataset

    @property
    def collection(self):
        return self.__collection

    @property
    def all(self):
        return self.__all



class JsonStatException(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return self.value


class JsonStatMalformedJson(JsonStatException):
    pass



from .exceptions import JsonStatException
from .exceptions import JsonStatMalformedJson

from .dimension import JsonStatDimension
from .value import JsonStatValue
from .dataset import JsonStatDataSet
from .collection import JsonStatCollection

import os
_examples_dir = os.path.abspath(os.path.join(
    os.path.dirname(__file__), "examples"))



def lst2html(lst):
    html = "<table>"
    for r in lst:
        html += "<tr>"
        for c in r:
            html += "<td>{}</td>".format(c)
        html += "</tr>"
    html += "</table>"
    return html



from collections import namedtuple
import json
import terminaltables
from .exceptions import JsonStatException, JsonStatMalformedJson

JsonStatCategory = namedtuple('JsonStatCategory', ['label', 'index', 'pos'])


class JsonStatDimension:
    def __init__(self, did=None, size=None, pos=None, role=None):
        """initialize a dimension

        .. warning::

            this is an internal library function (it is not public api)

        :param did: id of dimension
        :param size: size of dimension (nr of values)
        :param pos: position of dimension into the dataset
        :param role: of dimension
        """

        # it is valid if is correctly built (f.e. it was parsed correctly)
        self.__valid = False

        self.__did = did
        self.__size = size
        self.__role = role
        self.__pos = pos
        self.__label = None

        # if indexes are not present in json __idx2cat will be None
        # if labels  are not present in json __lbl2cat will be None
        self.__pos2cat = None  # int -> cat
        self.__idx2cat = None  # idx -> cat
        self.__lbl2cat = None  # lbl -> cat

    #
    # queries
    #   dimension properties

    @property
    def did(self):
        """id of this dimension"""
        return self.__did

    @property
    def label(self):
        """label of this dimension"""
        return self.__label

    @property
    def role(self):
        """role of this dimension (can be time, geo or metric)"""
        return self.__role

    @property
    def pos(self):
        """position of this dimension with respect to the data set to which this dimension belongs"""
        return self.__pos

    def __len__(self):
        """size of this dimension"""
        return self.__size

    def __to_list(self):
        lst = [["pos", "idx", "label"]]
        for cat in self.__pos2cat:
            idx = cat.index
            lbl = cat.label
            if idx is None:
                idx = ""
            if lbl is None:
                lbl = ""
            row = [str(cat.pos), "'" + idx + "'", "'" + lbl + "'"]
            row = list(map(lambda x: "" if x is None else x, row))
            lst.append(row)
        return lst

    def __str__(self):
        if self.__pos2cat is None:
            return ""

        lst = self.__to_list()

        table = terminaltables.AsciiTable(lst)
        out = table.table
        return out

    def __repr__(self):
        """used by ipython to make a better representation"""
        return self.__str__()

    def _repr_html_(self):
        lst = self.__to_list()
        html = "<table>"
        maxlines = 5
        nr_line = 0
        while nr_line < maxlines and nr_line < len(lst):
            l = lst[nr_line]
            html += "<tr>"
            for c in l:
                html += "<td>{}</td>".format(c)
            html += "</tr>"
            nr_line += 1
        if nr_line < len(lst):
            html += "<td>...</td>" * len(lst[0])
        html += "</table>"
        return html

    #
    # queries
    #   categories
    #

    def category(self, spec):
        """return JsonStatCategory according to spec

        :param spec: can be index (string) or label (string) or a position (integer)
        :returns: a JsonStatCategory
        """
        if not self.__valid:
            raise JsonStatException(
                "dimension '{}': is not initialized".format(self.__did))

        if isinstance(spec, int) and spec < len(self.__pos2cat):
            cat = self.__pos2cat[spec]
            return cat

        # try first indexes
        if spec in self.__idx2cat:
            cat = self.__idx2cat[spec]
            return cat

        if self.__lbl2cat is not None and spec in self.__lbl2cat:
            cat = self.__lbl2cat[spec]
            return cat

        raise JsonStatException(
            "dimension '{}': unknown index or label '{}'".format(self.__did, spec))

    def _pos2cat(self, pos):
        """get the category associated with the position (integer)

        :param pos: integer
        :returns: the label or None if the label not exists at position pos
            ex.: JsonStatCategory(index='2013', label='2013', pos=pos)
        """
        if not self.__valid:
            raise JsonStatException(
                "dimension '{}': is not initialized".format(self.__did))
        if self.__pos2cat is None:
            return None
        return self.__pos2cat[pos]

    def _idx2pos(self, idx):
        """from index to position

        :param idx: index for ex.: "2013"
        :returns: integer
        """
        if not self.__valid:
            raise JsonStatException(
                "dimension '{}': is not initialized".format(self.__did))
        if idx not in self.__idx2cat:
            raise JsonStatException(
                "dimension '{}': do not have index '{}'".format(self.__did, idx))
        return self.__idx2cat[idx].pos

    def _lbl2pos(self, lbl):
        """from label to position

        :param lbl: index for ex.: "2013"
        :returns: integer
        """
        if not self.__valid:
            raise JsonStatException(
                "dimension '{}': is not initialized".format(self.__did))
        if lbl not in self.__idx2cat:
            raise JsonStatException(
                "dimension '{}': do not have label {}".format(self.__did, lbl))
        return self.__lbl2cat[lbl].pos

    #
    # parsing methods
    #

    def from_string(self, json_string):
        """parse a json string

        :param json_string:
        :returns: itself to chain calls
        """
        json_data = json.loads(json_string)
        self.from_json(json_data)
        return self

    def from_json(self, json_data):
        """Parse a json structure representing a dimension

        From `json-stat.org <https://json-stat.org/format/#dimensionid>`_

            It is used to describe a particular dimension.
            The name of this object must be one of the strings in the id array.
            There must be one and only one dimension ID object for every dimension in the id array.

        jsonschema for dimension is about::

            "dimension": {
                "type": "object",
                "properties": {
                    "version": {"$ref": "#/definitions/version"},
                    "href": {"$ref": "#/definitions/href"},
                    "class": {"type": "string", "enum": ["dimension"]},
                    "label": {"type": "string"},
                    "category": {"$ref": "#/definitions/category"},
                    "note": {"type": "array"},
                },
                "additionalProperties": false
            },

        :param json_data:
        :returns: itself to chain call
        """
        # children category, label, class
        if 'label' in json_data:
            self.__label = json_data['label']

        if 'class' in json_data:
            if json_data['class'] != 'dimension':
                msg = "class must be equals to 'dimension'"
                raise JsonStatMalformedJson(msg)

        # parsing category
        if "category" not in json_data:
            msg = "dimension '{}': missing category key".format(self.__did)
            raise JsonStatMalformedJson(msg)

        self.__parse_category(json_data['category'])

        self.__valid = True
        return self

    def __parse_category(self, json_data_category):
        """It is used to describe the possible values of a dimension.
        See https://json-stat.org/format/#category
        :param json_data_category:
        :returns:

        jsonschema for dimension is about::

            "category": {
                "type": "object",
                "properties": {
                    "index": {"$ref": "#/definitions/category_index"},
                    "label": {"type": "object"},

                    "unit": {"$ref": "#/definitions/category_index"},
                    "child": {"type": "object", "properties": {"additionalProperties": {"type": "array"}}},
                    "coordinates": {"type": "object",
                                    "properties": {"additionalProperties": {"type": "array"}}},
                    "note": {"type": "array"}

                },
                "additionalProperties": false
            },

        """

        # validate: label or index must be present
        if 'index' not in json_data_category and 'label' not in json_data_category:
            msg = "dimension '{}': one of keys 'label' or 'index' must be presents"
            raise JsonStatMalformedJson(msg)

        if 'index' in json_data_category:
            self.__parse_json_index(json_data_category)

        if 'label' in json_data_category:
            self.__parse_json_label(json_data_category)

        # validate: number of indexes and labels must the same??
        if self.__idx2cat is not None and self.__lbl2cat is not None:
            if len(self.__idx2cat) != len(self.__lbl2cat):
                # TODO: cannot raise exception, emit warning see hierarchy.json
                msg = "dimension '{}': the number of indexes ({}) are different of the numbers of labels ({})"
                msg = msg.format(self.__did, len(
                    self.__idx2cat), len(self.__lbl2cat))
                # raise JsonStatMalformedJson(msg)
            if len(self.__idx2cat) < len(self.__lbl2cat):
                msg = "dimension '{}': the number of labels ({}) are greater than number of indexes ({})"
                msg = msg.format(self.__did, len(
                    self.__lbl2cat), len(self.__idx2cat))
                raise JsonStatMalformedJson(msg)

        # validate: indexes must be consistent with size
        if self.__size != len(self.__idx2cat):
            msg = "dimension '{}': malformed json: number of indexes {} not match with size {}"
            msg = msg.format(self.__did, len(self.__idx2cat), self.__size)
            raise JsonStatMalformedJson(msg)

        # validate: no hole in the indexes
        if any(v is None for v in self.__pos2cat):
            msg = "dimension '{}':hole in index".format(self.__did)
            raise JsonStatMalformedJson(msg)

        # "category_unit": {
        #                      "type": "object",
        #                      "properties": {
        #                          "additionalProperties": {
        #                              "type": "object",
        #                              "properties": {"label": {"type": "string"},
        #                                             "decimals": {"type": "number"},
        #                                             "type": {"type": "string"},
        #                                             "base": {"type": "string"},
        #                                             "multiplier": {"type": "number"},
        #                                             "position": {"type": "string"}},
        #                              "additionalProperties": false
        #                          }
        #                      }
        #                  },

        # TODO: parse 'unit'
        # "unit" : {
        # 	 "exp" : {
        # 			"decimals": 1,
        # 			"label" : "millions",
        # 			"symbol" : "$",
        # 			"position" : "start"
        # 	 }
        # }
        # 	"category" : {
        # 		"label" : {
        # 			"UNR" : "unemployment rate"
        # 		},
        # 		"unit" : {
        # 			"UNR" : {
        # 				"label" : "%",
        # 				"decimals" : 9,
        # 				"type" : "ratio",
        # 				"base" : "per cent",
        # 				"multiplier" : 0
        # 			}
        # 		}
        # 	}
        if 'unit' in json_data_category:
            if self.__role != "metric":
                msg = "dimension {}: 'unit' can be used only when role is 'metric'"
                msg = msg.format(self.__did)
                JsonStatException(msg)
            self.__unit = json_data_category['unit']

    def __parse_json_index(self, json_data):
        """parse index json structure

        for ex. the json structure could be
            "category" : { "index" : { "2003" : 0, "2004" : 1, "2005" : 2, "2006" : 3 }
        :param json_data: json structure
        """

        json_data_index = json_data['index']
        if self.__size is None:
            self.__size = len(json_data_index)

        self.__idx2cat = {}

        # preallocate a list of length self.size with default value None
        self.__pos2cat = self.__size * [None]

        if type(json_data_index) is list:
            for pos, idx in enumerate(json_data_index):
                self.__parse_json_index_helper(idx, pos)
        else:
            for idx, pos in json_data_index.items():
                self.__parse_json_index_helper(idx, pos)

    def __parse_json_index_helper(self, idx, pos):
        if pos >= self.__size:
            msg = "dimension '{}': index {} is greater than size {}"
            msg = msg.format(self.__did, pos, self.__size)
            raise JsonStatException(msg)

        cat = JsonStatCategory(pos=pos, index=idx, label=None)
        self.__pos2cat[pos] = cat
        self.__idx2cat[idx] = cat

    def __parse_json_label(self, json_data):
        """parse label structure

            "category" : {"label" : { "CA" : "Canada" }}
        :param json_data: json structure to parse
        """

        json_data_label = json_data['label']
        if self.__size is None:
            self.__size = len(json_data_label)

        no_index = 'index' not in json_data
        if no_index:
            # self.__size = len(json_data['label'])
            self.__pos2cat = self.__size * [None]
            self.__idx2cat = {}

        self.__lbl2cat = {}

        for i, (idx, lbl) in enumerate(json_data_label.items()):

            if no_index:
                # if index are not defined in json, give an order to the label
                pos = i
                cat = JsonStatCategory(pos=pos, label=lbl, index=idx)
            else:
                cat = self.__idx2cat.get(idx)
                if cat is None:
                    msg = "dimension '{}': label '{}' is associated with index '{}' that not exists!"
                    msg = msg.format(self.__did, lbl, idx)
                    raise JsonStatMalformedJson(msg)
                pos = cat.pos
                cat = JsonStatCategory(pos=pos, label=lbl, index=idx)

            self.__pos2cat[pos] = cat
            # if only labels are present into the json,  deduce indexes from labels
            self.__idx2cat[idx] = cat
            self.__lbl2cat[lbl] = cat



from functools import reduce
import json
import numpy as np
import pandas as pd
import terminaltables
from .value import JsonStatValue
from .dimension import JsonStatDimension
from .exceptions import JsonStatException, JsonStatMalformedJson
from .utility import lst2html


class JsonStatDataSet:
    def __init__(self, name=None):
        """Initialize an empty dataset.

        Dataset could have a name (key) if we parse a jsonstat format version 1.

        :param name: dataset name (for jsonstat v.1)
        """
        self.__valid = False

        self.__name = name
        self.__title = None
        self.__label = None
        self.__source = None

        # dimensions
        self.__dim_nr = 0  # len(self.__pos2dim)

        self.__pos2size = []  # array int -> int (dimension size)
        self.__pos2mult = None  # array int -> multiplicative factor

        self.__pos2dim = []  # array int -> dim
        self.__did2dim = {}  # dict  id  -> dim
        self.__lbl2dim = {}  # dict  lbl -> dim

        self.__value = None
        self.__status = None

    @property
    def name(self):
        """
        :getter: returns the name of the dataset
        :type: string
        """
        return self.__name

    @property
    def label(self):
        """
        :getter: returns the label of the dataset
        :type: string
        """
        return self.__label

    def __str__(self):
        out = ""
        if self.__name is not None:
            out += "name:   '{}'\n".format(self.__name)

        if self.__title is not None:
            out += "title:  '{}'\n".format(self.__title)

        if self.__label is not None:
            out += "label:  '{}'\n".format(self.__label)

        if self.__source is not None:
            out += "source: '{}'\n".format(self.__label, self.__source)

        out += "size: {}".format(len(self))
        out += "\n"
        out += self.__str__dimensions()
        return out

    def __repr__(self):
        """used by ipython to make a better representation"""
        return self.__str__()

    def _repr_html_(self):
        """used by ipython to make a better representation"""

        html = ""
        if self.__name is not None:
            html += "name:   '{}'</br>".format(self.__name)

        if self.__title is not None:
            html += "title:  '{}'</br>".format(self.__title)

        if self.__label is not None:
            html += "label:  '{}'</br>".format(self.__label)

        if self.__source is not None:
            html += "source: '{}'</br>".format(self.__label, self.__source)

        html += "size: {}</br>".format(len(self))
        lst = self.__dim_to_table()
        html += lst2html(lst)
        return html

    def __len__(self):
        """returns the size of the dataset"""
        return len(self.__value)

    #
    # dimensions
    #

    def dimensions(self):
        """returns list of JsonStatDimension"""
        return self.__pos2dim

    def dimension(self, spec):
        """get a JsonStatDimension by spec

        :param spec: spec can be:
         - (string) or id of the dimension
         - int position of dimension
        :returns: a JsonStatDimension
        """
        if type(spec) is int:
            return self.__pos2dim[spec]
        if spec not in self.__did2dim:
            msg = "dataset '{}': unknown dimension '{}' know dimensions ids are: {}"
            msg = msg.format(self.__name, spec, ", ".join(
                [dim.did for dim in self.__pos2dim]))
            raise JsonStatException(msg)
        return self.__did2dim[spec]

    def __dim_to_table(self):
        lst = [["pos", "id", "label", "size", "role"]]
        for i, dim in enumerate(self.__pos2dim):
            row = [str(i), dim.did, dim.label, str(len(dim)), dim.role]
            row = list(map(lambda x: "" if x is None else x, row))
            lst.append(row)
        return lst

    def __str__dimensions(self):
        lst = self.__dim_to_table()

        table = terminaltables.AsciiTable(lst)
        # table.justify_columns = {2: "right", 4: "right"}
        out = table.table
        return out

    def info_dimensions(self):
        """print same info on dimensions on stdout"""
        print(self.__str__dimensions())

    #
    # querying value/status
    #

    def data(self, *args, **kargs):
        """Returns a JsonStatValue containings value and status about a datapoint
        The datapoint will be retrieved according the parameters

        :param args:
            - data(<int>)  where i is index into the
            - data(<list>) where lst = [i1,i2,i3,...]) each i indicate the dimension len(lst) == number of dimension
            - data(<dict>) where dict is {k1:v1, k2:v2, ...} dimension of size 1 can be ommitted

        :param kargs:
            - data(k1=v1,k2=v2,...) where **ki** are the id or label of dimension
              **vi** are the index or label of the category dimension of size 1 can be ommitted

        :returns: a JsonStatValue object

        kargs { cat1:value1, ..., cati:valuei, ... }
        cati can be the id of the dimension or the label of dimension
        valuei can be the index or label of category
        ex.:{country:"AU", "year":"2014"}

        >>> import os, jsonstat  # doctest: +ELLIPSIS
        >>> filename = os.path.join(jsonstat._examples_dir, "www.json-stat.org", "oecd-canada-col.json")
        >>> dataset = jsonstat.from_file(filename).dataset(0)
        >>> dataset.data(0)
        JsonStatValue(idx=0, value=5.943826289, status=None)
        >>> dataset.data(concept='UNR', area='AU', year='2003')
        JsonStatValue(idx=0, value=5.943826289, status=None)
        >>> dataset.data(area='AU', year='2003')
        JsonStatValue(idx=0, value=5.943826289, status=None)
        >>> dataset.data({'area':'AU', 'year':'2003'})
        JsonStatValue(idx=0, value=5.943826289, status=None)
        """
        if not self.__valid:
            raise JsonStatException('dataset not initialized')

        try:
            idx = self._2idx(*args, **kargs)
            value = None
            for key, val in self.__value.items():
                if int(key) == int(idx.real):
                    value = val
                    break
        except Exception as e:
            print(kargs)
            raise e

        #
        # status
        #
        if self.__status is None:
            status = None
        elif isinstance(self.__status, str):
            status = self.__status
        elif isinstance(self.__status, list) and len(self.__status) == 1:
            status = self.__status[0]
        elif isinstance(self.__status, dict) and int(idx.real) not in self.__status:
            status = None
        else:
            status = self.__status[idx]
        return JsonStatValue(idx, value, status)

    def value(self, *args, **kargs):
        """get a value
        For the parameters see py:meth:`jsonstat.JsonStatDataSet.data`.

        :returns: value (typically a number)
        """
        # TODO: add onlyvalue=true to extract only the value
        return self.data(*args, **kargs).value

    def status(self, *args, **kargs):
        """get datapoint status

        For the parameters see py:meth:`jsonstat.JsonStatDataSet.data`.

        :returns: status (typically a string)
        """
        # TODO: add onlystatus=true to extract only the value?
        return self.data(*args, **kargs).status

    def __value_from_vec_pos(self, lst):
        """

        :param lst: [0,3,4]
        :returns: value at dimension [0,3,4]
        """
        return self.__value[self.lint_as_idx(lst)]

    #
    # dataset can be access using different type of indexes
    # simple index is integer.
    # ex. dataset.data(0)
    # other type of idexes
    # lint [ <int1>, <int2>, <int3> ...]
    # lcat [ <cat1>, <cat2>, ... ]
    # dcat [ <did1>:<cat1>, <did2>:<cat1>, ... ]

    # this functions are only for library internal usage
    #

    def _2idx(self, *args, **kargs):
        """convert args to integer index """

        if len(args) == 1:
            # data(int)
            if isinstance(args[0], int):
                return args[0]
            # data([i1,i2,i3])
            elif isinstance(args[0], list):
                idx = self.lint_as_idx(args[0])
                return idx
            # data({k1:v1, k2:v2})
            elif isinstance(args[0], dict):
                dims = args[0]
                apos = self.dcat_to_lint(dims)
                idx = self.lint_as_idx(apos)
                return idx
        elif len(args) == 0:
            # data(k1:v1, k2:v2)
            dims = kargs
            # print(dims)
            apos = self.dcat_to_lint(dims)
            # print(apos)
            idx = self.lint_as_idx(apos)
            # print(idx)
            return idx

        msg = "unexpected parameters"
        raise JsonStatException(msg)

    def dcat_to_lint(self, dims):
        """Transforms a dimension dict to dimension array

        ::

            {"country":"AU", "year":2014} -> [1,2,3]

        :param dims: keys are dimension (id or label), value are categories
             "country" is the id of dimension
             "AU" is the category of dimension
        :returns: a list of integer
        """
        apos = len(self.__pos2dim) * [0]
        for (cat, val) in dims.items():
            # key is id
            if cat in self.__did2dim:
                dim = self.__did2dim[cat]
            # key is label
            elif cat in self.__lbl2dim:
                dim = self.__lbl2dim[cat]
            # key is not id or label so raise error
            else:
                allowed_categories = ", ".join(
                    ["'{}'".format(dim.did) for dim in self.__pos2dim])
                msg = "dataset '{}': category '{}' don't exists allowed categories are: {}"
                msg = msg.format(self.__name, cat, allowed_categories)
                raise JsonStatException(msg)

            apos[dim.pos] = dim.category(val).pos
        return apos

    def lint_as_idx(self, lst):
        """from a list of position get a index into value array

        [1,2,3] -> 10
        :param lst: list of integer
        :returns: an integer index into values
        """
        s = np.array(self.__pos2mult)
        r = s * lst
        return np.sum(r)

    def idx_as_lint(self, idx):
        """ 10 -> [<int1>, <int2>, ...]
        """
        lint = self.__dim_nr * [0]
        i = len(self.__pos2size) - 1
        while idx > 0 and i != 0:
            lint[i] = idx % self.__pos2size[i]
            idx -= (lint[i] * self.__pos2mult[i])
            i -= 1
        return lint

    def idx_as_lcat(self, idx):
        lint = self.idx_as_lint(idx)
        lcat = self.lint_as_lcat(lint)
        return lcat

    def lint_as_lcat(self, lint, without_one_dimension=False):
        """transforms an array of int into an array of cat

        [0,3,4] -> ['dimension 1 index', 'dimension 2 label', 'dimension 3 label']

        :param lint:  [0,3,4]
        :returns: ['dimension 1 index', 'dimension 2 label', 'dimension 3 label']
        """
        lcat = []
        for pos, lint_pos in enumerate(lint):
            dim = self.__pos2dim[pos]
            if not (without_one_dimension and len(dim) == 1):
                cat = dim._pos2cat(lint_pos).index
                lcat.append(cat)
        return lcat

    def _lint_to_llbl(self, apos, without_one_dimension=False):
        """transforms on array of dim into an array of label

        :param apos:  [0,3,4]
        :returns: ['dimension 1 label or index', 'dimension 2 label  or index', 'dimension 3 label  or index']
        """

        # vec_idx = len(vec_pos) * [None]
        aidx = []
        for pos in range(len(apos)):
            dim = self.__pos2dim[pos]
            lbl = dim._pos2cat(apos[pos]).label
            if lbl is None:
                lbl = dim._pos2cat(apos[pos]).index

            # vec_idx[i] = lbl
            if not (without_one_dimension and len(dim) == 1):
                aidx.append(lbl)

        return aidx

    def _from_aidx_to_adim(self, ldid):
        """From a list of dimension name to a list of numerical dimension position

          F.e.
          ["year", "country"] -> [1,0]
          ["country", "year"] -> [0,1]

        :returns: list of number
        """
        return [self.__did2dim[did].pos for did in ldid]

    #
    # generators
    #

    def all_pos(self, blocked_dims={}, order=None):
        """all_pos doc

        :param blocked_dims:  {"year":2013, country:"IT"}
        :param order: order
        :returns:
        """

        nr_dim = len(self.__pos2dim)
        if order is not None:
            if len(order) != nr_dim:
                msg = "length of the order vector is different from number of dimension {}".format(
                    nr_dim)
                raise JsonStatException(msg)
            if not isinstance(order[1], int):
                order = [self.__did2dim[iid].pos for iid in order]

        vec_pos_blocked = nr_dim * [False]
        vec_pos = nr_dim * [0]

        for (cat, idx) in blocked_dims.items():
            d = self.dimension(cat)
            vec_pos_blocked[d.pos] = True
            vec_pos[d.pos] = d._idx2pos(idx)

        pos2size = self.__pos2size

        if order is None:
            vec_dimension_reorder = range(nr_dim)
        else:
            vec_dimension_reorder = order

        nrd = nr_dim - 1
        while nrd >= 0:

            yield list(vec_pos)  # make a shallow copy of vec_pos

            nrd = nr_dim - 1
            cur_dim = vec_dimension_reorder[nrd]
            # se la posizione non e bloccata allora puoi far andare avanti la cifra
            if not vec_pos_blocked[cur_dim]:
                vec_pos[cur_dim] += 1

            # se non si arrivati all'ultima dimensione
            # e se la dimensione corrente non e al massimo valore o se la dimensione corrente e bloccata
            while nrd >= 0 and \
                    (vec_pos[cur_dim] == pos2size[cur_dim] or vec_pos_blocked[cur_dim]):

                # se la posizione non e' bloccata allora puoi far partire il valore dall'inizio
                if not vec_pos_blocked[cur_dim]:
                    vec_pos[cur_dim] = 0

                # esamina la prossima posizione
                nrd -= 1
                # se la dimensione corrente non e' la prima
                if nrd >= 0:
                    cur_dim = vec_dimension_reorder[nrd]
                    # se la dimensione corrente non e bloccata puoi farla avanzare
                    if not vec_pos_blocked[cur_dim]:
                        vec_pos[cur_dim] += 1

    def generate_all_vec(self, **blocked_dims):
        for vec_pos in self.all_pos(blocked_dims):
            vec_idx = self.lint_as_lcat(vec_pos)
            value = self.__value_from_vec_pos(vec_pos)

    #
    # transforming function
    #

    def to_table(self,
                 content="label",
                 order=None,
                 rtype=list,
                 blocked_dims={},
                 value_column="Value",
                 without_one_dimensions=False):
        """Transforms a dataset into a table (a list of row)

        table len is the size of dataset + 1 for headers

        :param content: can be "label" or "id"
        :param order:
        :param rtype:
        :param blocked_dims:
        :returns: a list of row, first line is the header
        """
        table = []

        # header
        if content == "label":
            header = [dim.label for dim in self.__pos2dim]
        else:
            header = [dim.did for dim in self.__pos2dim]

        header.append(value_column)

        # data
        table.append(header)
        for apos in self.all_pos(order=order, blocked_dims=blocked_dims):
            value = self.__value_from_vec_pos(apos)
            if content == "label":
                row = self._lint_to_llbl(
                    apos, without_one_dimension=without_one_dimensions)
            else:
                row = self.lint_as_lcat(
                    apos, without_one_dimension=without_one_dimensions)
            row.append(value)
            table.append(row)

        if rtype == pd.DataFrame:
            ret = pd.DataFrame(table[1:], columns=table[0])
        else:
            ret = table

        return ret

    def to_data_frame(self,
                      index=None,
                      content="label",
                      order=None,
                      blocked_dims={},
                      value_column="Value"):
        """Transform dataset to pandas data frame

        extract_bidimensional("year", "country")
        generate the following dataframe:
        year  |  country
        2010  |  1
        2011  |  2
        2012  |  3

        :param index:
        :param content:
        :param blocked_dims:
        :param order:
        :param value_column:

        :returns:
        """

        df = self.to_table(content=content,
                           order=order,
                           rtype=pd.DataFrame,
                           blocked_dims=blocked_dims,
                           value_column=value_column)
        # TODO: avoid creating a new dataframe (?)
        # df.index = df[index]
        # del df[index]
        if index:
            df = df.set_index([index])
        return df

    #
    # Parsing code
    #

    def from_file(self, filename):
        """read a jsonstat from a file and parse it to initialize this dataset.

        It is better to use :py:meth:`jsonstat.from_file`

        :param filename: path of the file.
        :returns: itself to chain calls
        """
        with open(filename) as f:
            json_string = f.read()
            self.from_string(json_string)
        return self

    def from_string(self, json_string):
        """parse a string containing a jsonstat and initialize this dataset

        It is better to use :py:meth:`jsonstat.from_string`

        :param json_string: string containing a jsonstat
        :returns: itself to chain calls
        """
        json_data = json.loads(json_string)
        self.from_json(json_data)
        return self

    def from_json(self, json_data):
        """parse a json structure and initialize this dataset

        It is better to use py:meth:`jsonstat.from_json`

        :param json_data: json structure
        :returns: itself to chain calls
        """
        if "version" in json_data:
            # assume version 2
            self._from_json_v2(json_data)
        else:
            self._from_json_v1(json_data)
        return self

    def _from_json_v1(self, json_data):
        """parse a json structure according to jsonstat format version 1.x

        .. warning::

            this is an internal library function (it is not public api)

        :param json_data: json structure
        """

        if 'label' in json_data:
            self.__label = json_data['label']
            if self.__name is None:
                self.__name = self.__label

        if 'source' in json_data:
            self.__source = json_data['source']

        if 'title' in json_data:
            self.__title = json_data['title']

        # parsing value
        if 'value' not in json_data:
            msg = "dataset '{}': missing 'value' key".format(self.__name)
            raise JsonStatMalformedJson(msg)
        self.__value = json_data['value']

        if len(self.__value) == 0:
            msg = "dataset '{}': field 'value' is empty".format(self.__name)
            raise JsonStatMalformedJson(msg)

        # https://json-stat.org/format/#status
        # parsing status
        #
        # eurostat has the following structure for status
        # status : {
        #   'value' : { "": "" }
        #   'category' : { ... }
        # }

        if 'status' in json_data:
            self.__status = json_data['status']
            if isinstance(self.__status, list):
                if len(self.__status) != 1 and len(self.__status) != len(
                        self.__value):
                    msg = "dataset '{}': incorrect size of status fields"
                    raise JsonStatMalformedJson(msg)
            if isinstance(self.__status, dict):
                # convert key into int
                # eurostat data has incorrect status { "":"" }
                nd = {}
                for k, v in self.__status.items():
                    try:
                        nd[int(k)] = v
                    except ValueError:
                        pass
                self.__status = nd

        #
        # parsing dimension
        #
        if 'dimension' not in json_data:
            msg = "dataset '{}': missing 'dimension' key".format(self.__name)
            raise JsonStatMalformedJson(msg)

        json_data_dimension = json_data['dimension']

        if 'id' not in json_data_dimension:
            msg = "dataset '{}': missing 'dimension.id' key".format(
                self.__name)
            raise JsonStatMalformedJson(msg)

        if 'size' not in json_data_dimension:
            msg = "dataset '{}': missing 'dimension.size' key".format(
                self.__name)
            raise JsonStatMalformedJson(msg)

        pos2iid = json_data_dimension['id']

        self.__pos2size = json_data_dimension['size']
        # https://github.com/26fe/jsonstat.py/issues/1
        # cso.ie expose dimension sizes as strings instead of integers.
        for i, e in enumerate(self.__pos2size):
            self.__pos2size[i] = int(e)

        self.__dim_nr = len(pos2iid)

        # validate dimension
        if len(pos2iid) != len(self.__pos2size):
            msg = "dataset '{}': dataset_id is different of dataset_size".format(
                self.__name)
            raise JsonStatMalformedJson(msg)

        json_data_roles = None
        if 'role' in json_data_dimension:
            json_data_roles = json_data_dimension['role']
        self.__parse_dimensions(json_data_dimension, json_data_roles, pos2iid)

        # validate
        size_total = reduce(lambda x, y: x * y, self.__pos2size)
        if len(self.__value) != size_total:
            msg = "dataset '{}': size {} is different from calculate size {} by dimension"
            msg = msg.format(self.__name, len(self.__value), size_total)
            raise JsonStatMalformedJson(msg)

        self.__compute_pos2mult()
        self.__valid = True

    def _from_json_v2(self, json_data):
        """parse a jsonstat structure compliant to jsonstat format version 2.x

        .. warning::

            this is an internal library function (it is not public api)

        :param json_data: json structure

        keys to be parsed
        - version
        - class: "dataset"

        - href: url
        - label: "..."
        - id: <list of dimension id>
        - size: <list of integer, size of dimension>
        - role: roles of dimension
        - value: <list of values>
        - status
        - dimension
        - link

        ::

            {
                "class" : "dataset",
                "href" : "http://json-stat.org/samples/oecd.json",
                "label" : "Unemployment rate in the OECD countries 2003-2014"
             }

        """

        if "href" in json_data:
            self.__href = json_data["href"]

        if 'label' in json_data:
            self.__label = json_data['label']
            if self.__name is None:
                self.__name = self.__label

        if "id" not in json_data:
            if "href" in json_data:
                # todo skip the next section???
                # todo: download data?
                return

        # value is required
        # https://json-stat.org/format/#value
        # TODO: value into a numpy array?
        self.__value = json_data['value']
        if len(self.__value) == 0:
            msg = "dataset '{}': field 'value' is empty".format(self.__name)
            raise JsonStatMalformedJson(msg)

        pos2iid = json_data['id']

        self.__pos2size = json_data['size']
        # https://github.com/26fe/jsonstat.py/issues/1
        # cso.ie expose dimension sizes as strings instead of integers.
        for i, e in enumerate(self.__pos2size):
            self.__pos2size[i] = int(e)

        self.__dim_nr = len(pos2iid)

        # validate len(ids) == len(sizes)
        if len(pos2iid) != len(self.__pos2size):
            msg = "dataset '{}': dataset_id is different of dataset_size".format(
                self.__name)
            raise JsonStatMalformedJson(msg)

        # https://json-stat.org/format/#status
        # parsing status
        if 'status' in json_data:
            self.__status = json_data['status']
            if isinstance(self.__status, list):
                if len(self.__status) != 1 and len(self.__status) != len(
                        self.__value):
                    msg = "dataset '{}': incorrect size of status fields"
                    raise JsonStatMalformedJson(msg)
            if isinstance(self.__status, dict):
                # convert key into int
                self.__status = {int(k): v for k, v in self.__status.items()}

        # dimension
        json_data_roles = None
        if 'role' in json_data:
            json_data_roles = json_data['role']
        json_data_dimension = json_data["dimension"]
        self.__parse_dimensions(json_data_dimension, json_data_roles, pos2iid)

        # TODO: parsing link

        self.__compute_pos2mult()
        self.__valid = True

    def __parse_dimensions(self, json_data_dimension, json_data_roles,
                           pos2iid):
        """Parse dimension in json stat

        it used for format v1 and v2

        :param json_data_dimension:
        :param json_data_roles:
        :returns:
        """

        # parsing roles
        roles = {}
        if json_data_roles is not None:
            json_roles = json_data_roles
            for r in json_roles.items():
                role = r[0]
                for dname in r[1]:
                    roles[dname] = role

        # parsing each dimensions
        self.__pos2dim = self.__dim_nr * [None]
        for dpos, dname in enumerate(pos2iid):
            dsize = self.__pos2size[dpos]

            if dname not in json_data_dimension:
                msg = "dataset '{}': malformed json: missing key {} in dimension".format(
                    self.__name, dname)
                raise JsonStatException(msg)

            dimension = JsonStatDimension(dname, dsize, dpos, roles.get(dname))
            dimension.from_json(json_data_dimension[dname])
            self.__did2dim[dname] = dimension
            self.__pos2dim[dpos] = dimension
            if dimension.label is not None:
                self.__lbl2dim[dimension.label] = dimension

    def __compute_pos2mult(self):
        acc = 1
        self.__pos2mult = self.__dim_nr * [1]
        i = self.__dim_nr - 2
        while i >= 0:
            acc = acc * self.__pos2size[i + 1]
            self.__pos2mult[i] = acc
            i -= 1



from jsonstatpy import JsonStatDataSet
import pandas as pd
import json
import aiohttp


class EuroStatPy:
    def __init__(self, path: str = None) -> None:
        self.path = "./temp" if path is None else path
        self.base = "https://ec.europa.eu/eurostat/api/dissemination/statistics/1.0/data/"
        self.tail = "?format=JSON&lang=EN"
        self.since_time_period = "&sinceTimePeriod=2020"

    async def get_table_from_id(self, table_id: str) -> JsonStatDataSet:
        data = await self.__query_api(table_id)
        if data['class'] == 'dataset':
            jsonStatDataSet = JsonStatDataSet()
            jsonStatDataSet._from_json_v2(data)
            return jsonStatDataSet
        else:
            return f"API returned type {data['class']} which isn't currently supported by the package."

    async def get_table_from_name(self, table_name: str) -> JsonStatDataSet:
        id = self.__get_id_from_name(table_name)
        data = await self.__query_api(id)
        if data['class'] == 'dataset':
            jsonStatDataSet = JsonStatDataSet()
            jsonStatDataSet._from_json_v2(data)
            return jsonStatDataSet
        else:
            return f"API returned type {data['class']} which isn't currently supported by the package."

    def get_table_from_id_as_pandas(self, table_id: str) -> pd.DataFrame:
        jsd = self.get_table_from_id(table_id)
        return jsd.to_data_frame()

    def get_table_from_name_as_pandas(self, table_name: str) -> pd.DataFrame:
        jsd = self.get_table_from_name(table_name)
        return jsd.to_data_frame()

    @property
    def datasets(self) -> list[str]:
        raw_datasets = self.__get_datasets()
        return list(set(i[0].split('/')[-1] for i in raw_datasets))

    @property
    def codes(self) -> list[str]:
        raw_datasets = self.__get_datasets()
        return list(set(i[-1] for i in raw_datasets))

    def list_datasets(self) -> None:
        datasets = self.__get_datasets()
        idx = 1
        for ds in datasets:
            spds = ds[0].split('/')
            print(f"{idx}. Dataset: {spds[-1]} | Code: {ds[1]}")
            idx += 1

    async def __query_api(self, table_id: str, retry_url: str = None) -> dict:
        url = f"{self.base}{table_id}{self.tail}" if retry_url is None else retry_url
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                if response.status == 200:
                    return await response.json()
                elif response.status == 413:
                    await self.__query_api(table_id=table_id, retry_url=f"{self.base}{table_id}{self.tail}&sinceTimePeriod=2024")
                else:
                    raise Exception(
                        f"API returned with status code: {response.status}\n{await response.text()}")

    def __flatten_json(self, json_obj, parent_key='') -> list[tuple]:
        """
        Recursively flattens a nested JSON object into a list of tuples.

        Args:
        - json_obj (dict or list): The JSON object to flatten.
        - parent_key (str): The base key string to prefix the keys in the tuples.

        Returns:
        - List[Tuple[str, str]]: A list of tuples where each tuple is (key, value).
        """
        items = []
        if isinstance(json_obj, dict):
            for k, v in json_obj.items():
                new_key = f"{parent_key}/{k}" if parent_key else k
                if isinstance(v, dict) or isinstance(v, list):
                    items.extend(self.__flatten_json(v, new_key))
                else:
                    items.append((new_key, v))
        elif isinstance(json_obj, list):
            for i, item in enumerate(json_obj):
                new_key = f"{parent_key}/{i}" if parent_key else str(i)
                items.extend(self.__flatten_json(item, new_key))

        return items

    def __get_datasets(self):
        path = "eu_database.json"
        with open(path, 'r') as file:
            data = json.load(file)
        return self.__flatten_json(data)

    def __get_id_from_name(self, table_name: str) -> str:
        try:
            datasets = self.__get_datasets()
            # TODO: Fix this
            for k, v in datasets:
                if table_name in k:
                    return v
            raise Exception(f"Table name: {table_name} not found in dataset.")
        except Exception as e:
            raise e



from setuptools import setup

setup(
    name='eurostatpy',
    version='0.2.1',
    description='Basic package for querying & downloading EuroStat data by dataset name or ID.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=['pandas', 'terminaltables']
)



from collections import OrderedDict
import dateutil.parser
import json
import terminaltables
from .dataset import JsonStatDataSet
from .utility import lst2html


class JsonStatCollection:
    def __init__(self):
        self.__href = None
        self.__label = None
        self.__updated = None

        self.__name2dataset = {}  # str -> dataset
        self.__pos2dataset = []  # int -> dataset

    def __len__(self):
        """the number of dataset contained in this collection"""
        return len(self.__pos2dataset)

    def dataset(self, spec):
        """select a dataset belonging to the collection

        :param spec: can be

            - the name of collection (string) for jsonstat v1
            - an integer (for jsonstat v1 and v2)

        :returns: a JsonStatDataSet object
        """

        if type(spec) is int:
            return self.__pos2dataset[spec]
        return self.__name2dataset[spec]

    def __to_table(self):
        lst = [["pos", "dataset"]]
        for i, dataset in enumerate(self.__pos2dataset):
            row = [str(i), "'" + dataset.name + "'"]
            lst.append(row)
        return lst

    def __str__(self):
        out = "JsonstatCollection contains the following JsonStatDataSet:\n"
        lst = self.__to_table()
        table = terminaltables.AsciiTable(lst)
        out += table.table
        return out

    def __repr__(self):
        return self.__str__()

    def _repr_html_(self):
        """used by jupyter to make a better representation into notebooks"""
        html = "JsonstatCollection contains the following JsonStatDataSet:</br>"
        lst = self.__to_table()
        html += lst2html(lst)
        return html

    #
    # parsing methods
    #
    def from_file(self, filename):
        """initialize this collection from a file
        It is better to use :py:meth:`jsonstat.from_file`

        :param filename: name containing a jsonstat
        :returns: itself to chain call
        """
        with open(filename) as f:
            json_string = f.read()
            self.from_string(json_string)
        return self

    def from_string(self, json_string):
        """Initialize this collection from a string
        It is better to use :py:meth:`jsonstat.from_string`

        :param json_string: string containing a json
        :returns: itself to chain call
        """
        json_data = json.loads(json_string, object_pairs_hook=OrderedDict)
        self.from_json(json_data)
        return self

    def from_json(self, json_data):
        """initialize this collection from a json structure
        It is better to use :py:meth:`jsonstat.from_json`

        :param json_data: data structure (dictionary) representing a json
        :returns: itself to chain call
        """

        if "version" in json_data:
            self._from_json_v2(json_data)
        else:
            # jsonstat version 1.0
            self._from_json_v1(json_data)
        return self

    def _from_json_v1(self, json_data):
        """parse a jsonstat version 1

        .. warning::

            this is an internal library function (it is not public api)

        :param json_data: json structure

        json_data example::

            {
                "dataset1" : {...}
                "dataset2" : {...}
            }

        """
        for dataset_name, dataset_json in json_data.items():
            dataset = JsonStatDataSet(dataset_name)
            dataset.from_json(dataset_json)
            self.__name2dataset[dataset_name] = dataset
            self.__pos2dataset.append(dataset)

    def _from_json_v2(self, json_data):
        """parse a jsonstat version 2

        .. warning::

            this is an internal library function (it is not public api)

        :param json_data: json structure


        json_data example::

            "version" : "2.0",
            "class" : "collection",
            "href" : "http://json-stat.org/samples/collection.json",
            "label" : "JSON-stat Dataset Sample Collection",
            "updated" : "2015-07-02",
            "link" : {
               "item" : [
                    {},
                    {},
                ]
        }

        """

        if "href" in json_data:
            self.__href = json_data["href"]

        if "label" in json_data:
            self.__label = json_data["label"]

        if "updated" in json_data:
            self.__updated = dateutil.parser.parse(json_data["updated"])

        json_data_ds = json_data["link"]["item"]
        self.__pos2dataset = len(json_data_ds) * [None]
        for pos, ds in enumerate(json_data_ds):
            dataset = JsonStatDataSet()
            dataset._from_json_v2(ds)
            self.__pos2dataset[pos] = dataset
